@@ -2,10 +2,72 @@
 
 ;;; File: "_kernel#.scm"
 
-;;; Copyright (c) 1994-2018 by Marc Feeley, All Rights Reserved.
+;;; Copyright (c) 1994-2020 by Marc Feeley, All Rights Reserved.
 
 ;;;============================================================================
 
+;;; Representation of modules.
+
+(##define-macro (macro-make-module module-ref module-descr)
+  `(##vector ,module-ref 0 ,module-descr))
+
+(##define-macro (macro-module-module-ref module)
+  `(##vector-ref ,module 0))
+
+(##define-macro (macro-module-stage module)
+  `(##vector-ref ,module 1))
+
+(##define-macro (macro-module-stage-set! module x)
+  `(##vector-set! ,module 1 ,x))
+
+(##define-macro (macro-module-module-descr module)
+  `(##vector-ref ,module 2))
+
+(##define-macro (macro-make-module-descr
+                 supply-modules
+                 demand-modules
+                 meta-info
+                 flags
+                 thunk
+                 module-struct)
+  `(##vector ,supply-modules
+             ,demand-modules
+             ,meta-info
+             ,flags
+             ,thunk
+             ,module-struct))
+
+(##define-macro (macro-module-descr-supply-modules module-descr)
+  `(##vector-ref ,module-descr 0))
+
+(##define-macro (macro-module-descr-demand-modules module-descr)
+  `(##vector-ref ,module-descr 1))
+
+(##define-macro (macro-module-descr-meta-info module-descr)
+  `(##vector-ref ,module-descr 2))
+
+(##define-macro (macro-module-descr-flags module-descr)
+  `(##vector-ref ,module-descr 3))
+
+(##define-macro (macro-module-descr-flags-set! module-descr x)
+  `(##vector-set! ,module-descr 3 ,x))
+
+(##define-macro (macro-module-descr-thunk module-descr)
+  `(##vector-ref ,module-descr 4))
+
+(##define-macro (macro-module-descr-thunk-set! module-descr x)
+  `(##vector-set! ,module-descr 4 ,x))
+
+(##define-macro (macro-module-descr-module-struct module-descr)
+  `(##vector-ref ,module-descr 5))
+
+(##define-macro (macro-module-descr-module-struct-set! module-descr x)
+  `(##vector-set! ,module-descr 5 ,x))
+
+(##define-macro (macro-module-last-init-stage) 2)
+
+;;;----------------------------------------------------------------------------
+
 ;;; Representation of exceptions.
 
 (define-library-type exception
@@ -47,6 +109,16 @@
   (arguments unprintable: read-only: no-functional-setter:)
 )
 
+(define-library-type-of-exception wrong-number-of-values-exception
+  id: A090C430-D98F-431E-97EE-2846052FBBE0
+  constructor: #f
+  opaque:
+
+  (vals unprintable: read-only: no-functional-setter:)
+  (code unprintable: read-only: no-functional-setter:)
+  (rte  unprintable: read-only: no-functional-setter:)
+)
+
 (define-library-type-of-exception keyword-expected-exception
   id: 3fd6c57f-3c80-4436-a430-57ea4457c11e
   constructor: #f
@@ -115,7 +187,7 @@
 
   (procedure unprintable: read-only: no-functional-setter:)
   (arguments unprintable: read-only: no-functional-setter:)
-  (arg-num   unprintable: read-only: no-functional-setter:)
+  (arg-id    unprintable: read-only: no-functional-setter:)
   (type-id   unprintable: read-only: no-functional-setter:)
 )
 
@@ -139,6 +211,24 @@
   (arguments unprintable: read-only: no-functional-setter:)
 )
 
+(define-library-type-of-exception file-exists-exception
+  id: DD464B90-C0B2-437F-99AA-C6B411016D09
+  constructor: #f
+  opaque:
+
+  (procedure unprintable: read-only: no-functional-setter:)
+  (arguments unprintable: read-only: no-functional-setter:)
+)
+
+(define-library-type-of-exception permission-denied-exception
+  id: db6b7b55-594c-41e2-9268-05c6977db53e
+  constructor: #f
+  opaque:
+
+  (procedure unprintable: read-only: no-functional-setter:)
+  (arguments unprintable: read-only: no-functional-setter:)
+)
+
 (define-library-type-of-exception module-not-found-exception
   id: CA9CA020-600A-4516-AA78-CBE91EC8BE14
   constructor: #f
@@ -173,18 +263,18 @@
 (##define-macro (macro-debug-settings-error-quit)          2)
 (##define-macro (macro-debug-settings-error-shift)         5)
 
-(##define-macro (macro-debug-settings-repl-mask)           384)
-(##define-macro (macro-debug-settings-repl-ide)            0)
-(##define-macro (macro-debug-settings-repl-console)        1)
-(##define-macro (macro-debug-settings-repl-stdio)          2)
-(##define-macro (macro-debug-settings-repl-remote)         3)
+(##define-macro (macro-debug-settings-repl-mask)           896)
+(##define-macro (macro-debug-settings-repl-console)        0)
+(##define-macro (macro-debug-settings-repl-stdio)          1)
+(##define-macro (macro-debug-settings-repl-stdio-and-err)  2)
+(##define-macro (macro-debug-settings-repl-client)         3)
 (##define-macro (macro-debug-settings-repl-shift)          7)
 
-(##define-macro (macro-debug-settings-user-intr-mask)      1536)
+(##define-macro (macro-debug-settings-user-intr-mask)      3072)
 (##define-macro (macro-debug-settings-user-intr-repl)      0)
 (##define-macro (macro-debug-settings-user-intr-defer)     1)
 (##define-macro (macro-debug-settings-user-intr-quit)      2)
-(##define-macro (macro-debug-settings-user-intr-shift)     9)
+(##define-macro (macro-debug-settings-user-intr-shift)     10)
 
 (##define-macro (macro-debug-settings-level settings)
   `(##fxwraplogical-shift-right
@@ -218,6 +308,14 @@
 
 ;;;----------------------------------------------------------------------------
 
+;;; Module install mode.
+
+(##define-macro (macro-module-install-mode-ask-never)     0)
+(##define-macro (macro-module-install-mode-ask-when-repl) 1)
+(##define-macro (macro-module-install-mode-ask-always)    2)
+
+;;;----------------------------------------------------------------------------
+
 ;;; Terminal settings.
 
 (##define-macro (macro-terminal-settings-encoding-mask)    15)
